using System;
using System.Windows.Forms ;
using System.Drawing ;
using System.Drawing.Drawing2D ;
using System.ComponentModel ;
using System.Reflection ;
using System.Diagnostics ;
using ExtendedDataGrid.ColumnStyles.Events;



namespace ExtendedDataGrid.ColumnStyles
{
	/// <summary>
	/// DataGridButtonColumn Class.
	/// </summary>
	[DefaultEvent("ButtonClick"),DefaultProperty("Text")]
	public class DataGridButtonColumn: DataGridColumnStyleBase		
		
	{
		#region <<  PRIVATE FIELDS  >>

		private ExtendedDataGrid.GridButton btn=new ExtendedDataGrid.GridButton();
		private ImageList imageList=null;
		private int defaultImageIndex=-1;
		private bool defaultEnabled=true ;
		private CurrencyManager listManager=null;
		private Color backColor=SystemColors.Control ;
		private Color foreColor=SystemColors.ControlText ;
		private Font font=null;
		private System.Drawing.ContentAlignment imageAlign=ContentAlignment.MiddleCenter ;
		private System.Drawing.ContentAlignment textAlign =ContentAlignment.MiddleCenter ;
		private event ButtonSetCellStyleEventHandler setStyle;
		private ButtonClickEventHandler buttonClick;
		private string buttonText; 
		private ResizeType resizeType=ResizeType.Fixed ;
		private SparseType sparse=SparseType.CurrentRowOnly;
		private StringFormat fmt;
		private int rowNumber=-1;
		private bool buttonVisible=true;
		

		#endregion private fields 

		#region << CONSTRUCTORS >>

		public DataGridButtonColumn():this(null){}

		public DataGridButtonColumn(PropertyDescriptor prop):base(prop)
		{
			this.btn.Hide();
			fmt=new StringFormat();
			this.font=this.btn.Font ;
			fmt.Alignment=StringAlignment.Center  ;
			this.btn.Click+=new EventHandler(btn_Click);
		}


		#endregion constructors

		#region << PROPERTIES >>
		/// <summary>
		/// Specifies whether the button appears only in the current row or in all the cells of the column 
		/// </summary>
		[DefaultValue(SparseType.CurrentRowOnly ),
		Description("Specifies whether the button appears only in the current row or in all the cells of the column"),
		Category("Behaviour")]
		public SparseType Sparse
		{
			get{return this.sparse ;}
			set
			{
				this.sparse=value;
				this.Invalidate();
			}
		}


		/// <summary>
		/// This event raises every time the cells in this column should paint 
		/// </summary>
		[Description("This event raises every time the cells in this column should paint")]
		public event ButtonSetCellStyleEventHandler SetCellStyle
		{
			add
			{
				this.setStyle+= value;
			}
			remove
			{
				this.setStyle-=value;
			}
		}


		/// <summary>
		/// This event raises when ths button is clicked 
		/// </summary>
		[Description("The Click event of the button inside the column")]
		public event ButtonClickEventHandler  ButtonClick
		{
			add{this.buttonClick+=value;}
			remove
			{
				this.buttonClick-=value;
			}
		}


		[Description("The property Text of the button"),DefaultValue(""),Category("Layout")]
		public string Text
		{
			get{return this.buttonText;}
			set{this.buttonText=value;}
		}
		

		[Browsable(false),EditorBrowsable(EditorBrowsableState.Never )]
		public override bool ReadOnly
		{
			get
			{
				return base.ReadOnly;
			}
			set
			{
				base.ReadOnly = value;
			}
		}

		[Browsable(false),EditorBrowsable(EditorBrowsableState.Never )]
		public override string NullText
		{//just to hide it
			get{return base.NullText ;}
		}


		[Description("Specifies the way the column will resize when the grid resizes"),
		DefaultValue(ResizeType.Fixed ),Category("Behaviour")]
		public override ResizeType ResizeType
		{
			get{return this.resizeType ;}
			set{this.resizeType=value;}
		}


		[Category("Layout")]
		public Color DefaultBackColor
		{get{return this.backColor;}   set{this.backColor=value;}	}
		public bool ShouldSerializeDefaultBackColor()
		{
		return this.backColor!=SystemColors.Control;
		}
		public void ResetDefaultBackColor()
		{
		this.backColor=SystemColors.Control ;
		}
		

		[Category("Layout")]
		public Color DefaultForeColor
		{get{return this.foreColor ;}  set{this.foreColor=value;}}

		public bool ShouldSerializeDefaultForeColor()
		{
		return (this.foreColor!=SystemColors.ControlText);
		}
		public void ResetDefaultForeColor()
		{
		this.foreColor=SystemColors.ControlText ;
		}


		[DefaultValue(null),RefreshProperties(RefreshProperties.Repaint),
		Category("Layout")]
		public ImageList ImageList 
		{get{return this.imageList;}   
			set{
				this.imageList=value;
				this.btn.ImageList=value;
				if (value==null)
					this.defaultImageIndex=-1;
			}
		}
		public bool ShouldSerializeImageList()
		{
		return this.imageList!=null;
		}
		public void ResetImageList()
		{
		this.imageList=null;
		}
		
		[Description("The default index of the image to show in the cells of this column"),
		DefaultValue(-1),
		Category("Layout"),
		TypeConverter(typeof(ImageIndexConverter)),System.ComponentModel.EditorAttribute("System.Windows.Forms.Design.ImageIndexEditor",typeof(System.Drawing.Design.UITypeEditor))]
		public int DefaultImageIndex
		{get{return this.defaultImageIndex;}   set{this.defaultImageIndex=value;this.btn.ImageIndex=value;}}


		[DefaultValue(null),Description("The background image for the button inside the cell"),
		Category("Layout")]
		public Image BackgroundImage
		{get{return this.btn.BackgroundImage  ;}   
		set
		{
			this.btn.BackgroundImage =value;
			this.btn.Refresh();
			this.Invalidate();

		}
		}
		public bool ShouldSerializeBackgroundImage()
		{
		return this.btn.BackgroundImage!=null;
		}
		public void ResetBackgroundImage()
		{
		this.btn.BackgroundImage=null;
		}

		private new HorizontalAlignment Alignment
		{
			// just for hiding it 
			get{return HorizontalAlignment.Left;}
		}

		[DefaultValue(true),Category("Behaviour")]
		public bool DefaultEnabled
		{get{return this.defaultEnabled ;}   
			set
			{
				this.defaultEnabled=value;
				if(this.btn.Visible )
					this.btn.Enabled=value;
			}
		}

		[Category("Behaviour"),DefaultValue(true)]
		public bool DefaultVisible
		{
			get{return this.buttonVisible ;}
			set{this.buttonVisible=value;}
		}


		[Category("Layout")]
		public Font DefaultFont 
		{get{return this.font;}   set{this.font=value;}}
		public bool ShouldSerializeDefaultFont()
		{
		return this.font!=this.btn.Font;
		}
		public void ResetDefaultFont()
		{this.font=this.btn.Font;}


		[Category("Layout"),DefaultValue(ContentAlignment.MiddleCenter )]
		public ContentAlignment ImageAlign
		{get{return this.imageAlign;}    set{this.imageAlign=value;this.btn.ImageAlign =value;}}


		[Category("Layout"),DefaultValue(ContentAlignment.MiddleCenter )]
		public ContentAlignment TextAlign
		{get{return this.textAlign;}    set{this.textAlign=value;this.btn.TextAlign=value;SetStringFormat();}}
		

		#endregion properties 

		#region << OVERRIDINGS >>
		protected override void SetDataGridInColumn(System.Windows.Forms.DataGrid value)
		{
			if(this.btn.Parent!=null)
				this.btn.Parent.Controls.Remove(this.btn);
			if (value!=null)
				value.Controls.Add(this.btn );
			
		}


		protected override void ReleaseHostedControl()
		{
			base.ReleaseHostedControl();
			if(this.btn.Parent!=null)
				this.btn.Parent.Controls.Remove(this.btn);
		}



		protected override void Abort(int rowNum)
		{
			//Debug.WriteLine("Abort rowNum="+rowNum.ToString());
			this.btn.Hide(); 
			this.Invalidate();
		}
		protected override bool Commit(CurrencyManager dataSource, int rowNum)
		{
			//Debug.WriteLine("Commit rowNum="+rowNum.ToString());
			this.btn.Hide();
			this.Invalidate();
			return true;
		}

		internal override void ReleaseFocus()
		{
			base.ReleaseFocus ();
			this.ConcedeFocus();
		}
		protected override void ConcedeFocus()
		{
			this.btn.Hide(); 
			base.ConcedeFocus ();
		}

		protected override void Edit(CurrencyManager source, int rowNum, Rectangle bounds, bool readOnly)
		{
			this.Edit (source, rowNum, bounds, readOnly,"",true);
		}
		protected override void Edit(CurrencyManager source, int rowNum, Rectangle bounds, bool readOnly, string instantText)
		{
			this.Edit (source, rowNum, bounds, readOnly, instantText,true);
		}
		

		protected  override void Edit(CurrencyManager source, int rowNum, Rectangle bounds, bool readOnly, string instantText, bool cellIsVisible)
		{
			base.Edit(source,rowNum,bounds,readOnly,instantText,cellIsVisible);
			this.listManager=source;
 
			if(cellIsVisible)
			{
				ButtonSetCellStyleEventArgs e =new ButtonSetCellStyleEventArgs( this.buttonText ,this.backColor,this.foreColor,this.font ,this.defaultImageIndex ,rowNum,true,this.defaultEnabled,this.buttonVisible  );
				OnSetCellStyle(e);

				if(e.CurrentVisible)
				{

					if(this.btn.BackgroundImage!=null)
					{
						this.btn.BackColor=e.CurrentBackColor ;
					}
					this.rowNumber=rowNum;
					this.btn.Font=e.CurrentFont ;
					this.btn.BackColor= e.CurrentBackColor ;
					this.btn.ForeColor=e.CurrentForeColor;
					this.btn.ImageIndex=e.CurrentImageIndex ;
					this.btn.Text=e.CurrentText ;
					this.btn.Font=e.CurrentFont ;
					this.btn.Bounds=bounds;
					this.btn.Enabled=e.CurrentEnabled;
					this.btn.Visible=true;
					this.btn.Select();
				}
			}
		
		}
		

//		protected override int GetMinimumHeight()
//		{
//			return this.btn.Height;
//		}
//
//		protected override int GetPreferredHeight(Graphics g, object value)
//		{
//			return this.btn.Height;
//		}
//
//		protected override Size GetPreferredSize(Graphics g, object value)
//		{
//			return this.btn.Size ;
//		}

		protected override void Paint(Graphics g, Rectangle bounds, CurrencyManager source, int rowNum)
		{
			this.Paint(g,bounds,source,rowNum,false );
		}

		protected override void Paint(Graphics g, Rectangle bounds, CurrencyManager source, int rowNum, bool alignToRight)
		{
			this.Paint(g,bounds,source,rowNum,null ,null,alignToRight );
		}

		protected override void Paint(Graphics g, Rectangle bounds, CurrencyManager source, int rowNum, Brush backBrush, Brush foreBrush, bool alignToRight)
		{
			Color bkColor;
			if (rowNum % 2==0)
				bkColor=this.DataGridTableStyle.DataGrid.BackColor ;
			else
				bkColor=this.DataGridTableStyle.DataGrid.AlternatingBackColor ;

			if((rowNum!=source.Position) && (this.sparse==SparseType.CurrentRowOnly)) 
			{
				g.FillRectangle(new SolidBrush(bkColor),bounds);
			}
			else
			{
				ButtonSetCellStyleEventArgs e =new ButtonSetCellStyleEventArgs(this.buttonText ,this.backColor,this.foreColor,this.font ,this.defaultImageIndex ,rowNum,false,this.defaultEnabled ,this.buttonVisible );
				OnSetCellStyle(e);

				if(e.CurrentVisible==false)
					g.FillRectangle(new SolidBrush(bkColor),bounds);
				else
					DrawButton(g,e,bounds);
			}
		}
 
		#endregion overridings 

		#region << Private/Protected  methods >>

		private void DrawButton(Graphics g, ButtonSetCellStyleEventArgs e,Rectangle bounds)
		{
		
			this.BeginUpdate();

			// background
			if(this.btn.BackgroundImage!=null)
				this.DrawBkgroundImage(g,bounds);
			else
				g.FillRectangle(new SolidBrush(e.CurrentBackColor),bounds);

			// border
			ControlPaint.DrawBorder3D(g,bounds,Border3DStyle.Raised );

			// client rect
			bounds.Inflate(-4,-4);

			// the rectangle of the image 
			Rectangle rectImage=Rectangle.Empty ;
			if(this.imageList!=null && e.CurrentImageIndex>-1 && e.CurrentImageIndex<this.imageList.Images.Count )
				rectImage=GetRectImage(bounds );

			//the rectangle for the Text
			Rectangle rectText=Rectangle.Empty ;
			if(e.CurrentText!="")
				rectText=GetRectText(g,e.CurrentFont,e.CurrentText,bounds);

			Image img=null;

			// draw the image 
			if (!rectImage.IsEmpty)
			{
				img=this.imageList.Images[e.CurrentImageIndex];

				if(e.CurrentEnabled==false )
					ControlPaint.DrawImageDisabled(g,img,rectImage.X,rectImage.Y,e.CurrentBackColor);
				else
					g.DrawImage(img,rectImage);
			
			}

			// draw the text
			if(!rectText.IsEmpty )
			{
				if(e.CurrentEnabled )
					g.DrawString(e.CurrentText,e.CurrentFont,new SolidBrush(e.CurrentForeColor ),rectText,this.fmt);
				else
					ControlPaint.DrawStringDisabled(g,e.CurrentText,e.CurrentFont,e.CurrentBackColor ,rectText,this.fmt);

			}	

			this.EndUpdate();
		}

		Rectangle GetRectImage(Rectangle bnd)
		{
			// gets the rectangle where the image will be drawn

			Size imgSize=this.imageList.ImageSize ;
			Rectangle result=Rectangle.Empty ;
			switch(this.imageAlign )
			{
				case ContentAlignment.MiddleCenter :
					result=new Rectangle(new Point(this.CenterX(bnd,imgSize),this.CenterY(bnd,imgSize)), imgSize);
					break;
				case ContentAlignment.MiddleLeft :
					result=new Rectangle(new Point(bnd.X,this.CenterY(bnd,imgSize)), imgSize);
					break;
				case ContentAlignment.MiddleRight:
					result=new Rectangle(new Point(Math.Max(bnd.Right-imgSize.Width,bnd.X) ,this.CenterY(bnd,imgSize)), imgSize);
					break;
				case ContentAlignment.TopCenter :
					result=new Rectangle(new Point(this.CenterX(bnd,imgSize),bnd.Top), imgSize);
					break;
				case ContentAlignment.TopLeft :
					result=new Rectangle(new Point(bnd.X ,bnd.Top), imgSize);
					break;
				case ContentAlignment.TopRight :
					result=new Rectangle(new Point(Math.Max(bnd.Right-imgSize.Width,bnd.X) ,bnd.Top), imgSize);
					break;
				case ContentAlignment.BottomCenter :
					result=new Rectangle(new Point(this.CenterX(bnd,imgSize) ,Math.Max(bnd.Bottom-imgSize.Height,bnd.Y) ), imgSize);
					break;
				case ContentAlignment.BottomLeft :
					result=new Rectangle(new Point(bnd.X ,Math.Max(bnd.Bottom-imgSize.Height,bnd.Y) ), imgSize);
					break;
				case ContentAlignment.BottomRight :
					result=new Rectangle(new Point(Math.Max( bnd.Right-imgSize.Width,bnd.X) ,Math.Max(bnd.Bottom-imgSize.Height,bnd.Y) ), imgSize);
					break;
			}
		return result;
	}


		int CenterX(Rectangle bnd,Size size)
		{
		int result;
		if (size.Width>=bnd.Width)
			result= bnd.X;
		else
			result= bnd.X+(bnd.Width-size.Width)/2;
		return result;
		}


		int CenterY(Rectangle bnd,Size size)
		{
		int result;
		if (size.Height>=bnd.Height )
			result= bnd.Y;
		else
			result= bnd.Y+(bnd.Height-size.Height )/2;
		return result;
		}
		

		void SetStringFormat()
		{
			if (this.textAlign==ContentAlignment.MiddleLeft  || this.textAlign==ContentAlignment.TopLeft  || this.textAlign==ContentAlignment.BottomLeft)
				this.fmt.Alignment=StringAlignment.Near ;
			else 
				if(this.textAlign==ContentAlignment.MiddleCenter  || this.textAlign==ContentAlignment.TopCenter  || this.textAlign==ContentAlignment.BottomCenter)
				this.fmt.Alignment=StringAlignment.Center;
			else
				this.fmt.Alignment=StringAlignment.Far ;

		}


		Rectangle GetRectText(Graphics g, Font font,string text,Rectangle bnd)
		{
			Size txtSize=g.MeasureString(text,font,bnd.Width,this.fmt ).ToSize();
			Rectangle result=Rectangle.Empty ;
			switch(this.textAlign)
			{
				case ContentAlignment.MiddleCenter:
				case ContentAlignment.MiddleLeft:
				case ContentAlignment.MiddleRight:
					if (txtSize.Height>=bnd.Height)
						result= bnd ;
					else
						result= new Rectangle(new Point(bnd.X,bnd.Y+(bnd.Height-txtSize.Height)/2),bnd.Size );
					break;
				case ContentAlignment.TopCenter:
				case ContentAlignment.TopLeft:
				case ContentAlignment.TopRight:
					result= new Rectangle(new Point(bnd.X,bnd.Y),bnd.Size );
					break;

				case ContentAlignment.BottomCenter:
				case ContentAlignment.BottomLeft:
				case ContentAlignment.BottomRight:
					result=  new Rectangle(new Point(bnd.X,Math.Max(bnd.Bottom-txtSize.Height,bnd.Y)),new Size(bnd.Width,Math.Min(txtSize.Height,bnd.Height )) );
					break;
			}
		return result;
		}
		

		void DrawBkgroundImage(Graphics g,Rectangle bnd)
		{
			// draw the background image
		TextureBrush backBrush=new TextureBrush(this.btn.BackgroundImage);
		g.FillRectangle(backBrush,bnd);
		backBrush.Dispose();
		}


		protected void OnSetCellStyle(ButtonSetCellStyleEventArgs e)
		{
			// generate the event SetCellStyle

			if(this.setStyle!=null)
			{
				this.setStyle(this,e);
			}
		}
		

		Size GetMinSize(Size size1,Size size2)
		{
		return new Size(Math.Min(size1.Width,size2.Width ),Math.Min(size1.Height,size2.Height ));
		}

		private void btn_Click(object sender, EventArgs e)
		{
			if(this.listManager!=null)
			{
				int rowIndex =Util.GetRowNumber(this.rowNumber,this.listManager);
				ButtonClickEventArgs eArg=new ButtonClickEventArgs(rowIndex );
				OnButtonClick(eArg);
				if (eArg.RowHasBeenModified) 
					this.ColumnStartedEditing(this.btn);

			}
			
		}
		protected virtual void OnButtonClick(ButtonClickEventArgs e)
		{
			if(this.buttonClick!=null)
			{
					this.buttonClick(this,e);
					this.Edit(null,this.rowNumber,this.btn.Bounds,false);
			}
		}
		#endregion Private/Protected methods


	} 



	/// *********************************************************************************
	///  EventHandlers , EventArgs , structures
	/// *********************************************************************************
	namespace Events
	{

	#region << ButtonSetCellStyle(Delegate + EventArgs) >> 
	public delegate void ButtonSetCellStyleEventHandler(object sender, ButtonSetCellStyleEventArgs e);

	[Description("The EventArgs for the event  SetCellStyle")]
	public class ButtonSetCellStyleEventArgs: EventArgs
	{
		#region << Private Fields >>

		private Color currentBackColor;
		private Color currentForeColor;
		private Font currentFont;
		private int currentImageIndex;
		private int currentRowNumber;
		private bool currentEnabled;
		private string currentText;
		private bool currentVisible;
		private bool isInEditing;

		#endregion Private Fields 

		public ButtonSetCellStyleEventArgs(string curText ,Color curBkColor,Color curForeColor,Font curFont,int curImageIndex,int curRowIndex,bool inEditing,bool curEnabled,bool curVisible)
		{
			this.currentBackColor=curBkColor;this.currentForeColor=curForeColor;
			this.currentFont=curFont;this.currentImageIndex=curImageIndex;this.currentRowNumber=curRowIndex;
			this.currentVisible=curVisible;
			this.isInEditing=inEditing;this.currentEnabled=curEnabled;this.currentText=curText;
		}

		#region << Properties >>

		/// <summary>
		/// Shows/Hides the button in the cell for which this event has been raised 
		/// </summary>
		public bool CurrentVisible
		{
			get{return this.currentVisible;}
			set{this.currentVisible=value;}
		}

		/// <summary>
		/// ReadOnly 
		/// Indicate whether the cell(for which this event was raised) is being edited 
		/// </summary>
		public bool IsInEditing
		{
			get{return this.isInEditing;}
			set{this.isInEditing=value;}
		}

		/// <summary>
		/// Gets/Sets the Text property of the button  in the cell for which this event has been raised
		/// </summary>
		public string CurrentText 
		{
			get{return this.currentText;}
			set{this.currentText=value;}
		}

		/// <summary>
		/// Gets/Sets the Enabled property of the button  in the cell for which this event has been raised
		/// </summary>
		public bool CurrentEnabled
		{
			get{return this.currentEnabled;}
			set{this.currentEnabled=value;}
		}

		/// <summary>
		/// Gets/Sets the BackColor of the button in the cell for which this event has been raised
		/// </summary>
		public Color CurrentBackColor
		{
			get{return this.currentBackColor;}
			set{this.currentBackColor=value;}
		}


		/// <summary>
		/// Gets/Sets the ForeColor of the button in the cell for which this event has been raised
		/// </summary>
		public Color CurrentForeColor
		{
			get{return this.currentForeColor ;}
			set{this.currentForeColor=value;}
		}

		/// <summary>
		/// Gets/Sets the Font of the button in the cell for which this event has been raised
		/// </summary>
		public Font CurrentFont
		{
			get{return this.currentFont;}
			set{this.currentFont=value;}
		}

		/// <summary>
		/// Gets/Sets the index of the image to show on the button in the cell for which this event has been raised
		/// </summary>
		public int CurrentImageIndex
		{
			get{return this.currentImageIndex ;}
			set{this.currentImageIndex=value;}
		}

		/// <summary>
		/// ReadOnly 
		/// Gets the index of the row for which this event has been raised  
		/// </summary>
		public int CurrentRowNumber
		{
			get{return this.currentRowNumber;}
		}
		#endregion Properties
	}
	#endregion SetCellStyle

	#region << ButtonClick (Delegate + EventArgs) >>
		public delegate void ButtonClickEventHandler(object sender,ButtonClickEventArgs e);

		public class ButtonClickEventArgs:EventArgs
		{
			private int currentRowNumber ;
			private bool rowHasBeenChanged=false;
		
			public ButtonClickEventArgs(int curRowNumber)
			{
				this.currentRowNumber=curRowNumber;
			}
		
			/// <summary>
			/// The index of the row for which this event has been raised
			/// </summary>
			public int CurrentRowNumber
			{
				get{return this.currentRowNumber;}
			}
			/// <summary>
			/// Set this property to true to indicate that you modified the current row.
			/// </summary>
			public bool RowHasBeenModified
			{
				get {return this.rowHasBeenChanged;}
				set{this.rowHasBeenChanged=value;}
			}
		}
		#endregion Button Click 

	}// end namespace Events

	#region    <<     Structures    >>
	public enum SparseType
	{CurrentRowOnly, AllCells}
	#endregion 

}
